@@ -1,4 +1,4 @@
-# $Id: Engine.py.in 3128 2010-04-01 12:58:50Z matthijs $
+# $Id: Engine.py.in 3653 2010-08-05 14:04:51Z matthijs $
 #
 # Copyright (c) 2009 NLNet Labs. All rights reserved.
 #
@@ -431,6 +431,7 @@ class Engine:
                         response += self.read_zonelist()
                         response += "\n" + self.check_zone_conf_updates()
                 self.notify_all()
+		self.notify_zonefetcher ()
             if command[:4] == "stop":
                 self.stop_engine()
                 response = "Engine stopped"
@@ -515,6 +516,21 @@ class Engine:
         syslog.syslog(syslog.LOG_INFO, "close syslog")
         syslog.closelog()
 
+    def notify_zonefetcher(self):
+        """Reload the zone list into the zone fetcher daemon"""
+        pid_file = self.config.zonefetch_pidfile
+        if not os.path.exists(pid_file):
+            syslog.syslog(syslog.LOG_ERR, "could not notify zone fetcher: pid file does not exist: " + pid_file)
+        else:
+            pf = open(pid_file, 'r')
+            pid = pf.read()
+            syslog.syslog(syslog.LOG_DEBUG, "send notify to pid " + pid)
+            pf.close()
+            try:
+                os.kill(int(pid), signal.SIGUSR1)
+            except OSError:
+                syslog.syslog(syslog.LOG_ERR, "could not notify zone fetcher: no such process " + pid)
+
     def read_zonelist(self):
         """Reads the list of zones from the zone list xml file. Added
         zones are automatically scheduled for signing at the appropriate
@@ -893,6 +909,6 @@ def parse_args():
 if __name__ == '__main__':
     result = parse_args()
     if result == 1:
-        print "OpenDNSSEC signer engine version 1.1.1"
+        print "OpenDNSSEC signer engine version 1.1.3"
         main()
 
