@@ -1,5 +1,5 @@
 /*
- * $Id: zone_fetcher.c 3027 2010-03-12 09:28:44Z matthijs $
+ * $Id: zone_fetcher.c 3912 2010-09-10 12:35:26Z rb $
  *
  * Copyright (c) 2009 NLnet Labs. All rights reserved.
  *
@@ -37,6 +37,7 @@
 #include <getopt.h>
 #include <signal.h>
 #include <syslog.h>
+#include <unistd.h>
 
 #include <libxml/tree.h>
 #include <libxml/parser.h>
@@ -49,6 +50,8 @@
 #define DNS_SERIAL_GT(a, b) ((int)(((a) - (b)) & 0xFFFFFFFF) > 0)
 
 static int sig_quit = 0;
+static int sig_reload = 0;
+static char* zonelist_file = NULL;
 
 static void
 usage(FILE *out)
@@ -67,6 +70,58 @@ usage(FILE *out)
     fprintf(out, "-z <file>     The zonelist.xml <file>\n");
 }
 
+ldns_resolver*
+init_xfrd(config_type* config)
+{
+    serverlist_type* servers;
+    ldns_rdf* ns = NULL;
+    ldns_status status = LDNS_STATUS_OK;
+
+    ldns_resolver* xfrd = ldns_resolver_new();
+    if (config) {
+        if (config->use_tsig) {
+            ldns_resolver_set_tsig_keyname(xfrd, config->tsig_name);
+            if (strncmp(config->tsig_algo, "hmac-md5", 8) == 0) {
+                ldns_resolver_set_tsig_algorithm(xfrd, "hmac-md5.sig-alg.reg.int.");
+            } else {
+                ldns_resolver_set_tsig_algorithm(xfrd, config->tsig_algo);
+            }
+            ldns_resolver_set_tsig_keydata(xfrd, config->tsig_secret);
+        }
+        if (config->serverlist && config->serverlist->port)
+            ldns_resolver_set_port(xfrd, atoi(config->serverlist->port));
+        else
+            ldns_resolver_set_port(xfrd, atoi(DNS_PORT_STRING));
+        ldns_resolver_set_recursive(xfrd, 0);
+
+        servers = config->serverlist;
+        while (servers) {
+            if (servers->family == AF_INET6)
+                ns = ldns_rdf_new_frm_str(LDNS_RDF_TYPE_AAAA, servers->ipaddr);
+            else
+                ns = ldns_rdf_new_frm_str(LDNS_RDF_TYPE_A, servers->ipaddr);
+            if (ns)
+                status = ldns_resolver_push_nameserver(xfrd, ns);
+            else {
+                log_msg(LOG_ERR, "zone fetcher could not use %s for transfer "
+                    "request: could not parse ip address", servers->ipaddr);
+            }
+            if (status != LDNS_STATUS_OK) {
+                log_msg(LOG_ERR, "zone fetcher could not use %s for transfer "
+                    "request: %s", servers->ipaddr,
+                    ldns_get_errorstr_by_id(status));
+            }
+            servers = servers->next;
+        }
+        if (ldns_resolver_nameserver_count(xfrd) <= 0) {
+            log_msg(LOG_ERR, "zone fetcher could not find any valid name "
+                "servers");
+        }
+
+    }
+    return xfrd;
+}
+
 static zonelist_type*
 new_zone(char* zone_name, char* input_file)
 {
@@ -81,14 +136,15 @@ new_zone(char* zone_name, char* input_file)
 static void
 free_zonelist(zonelist_type* zlt)
 {
-    if (zlt) {
-        free_zonelist(zlt->next);
+    while (zlt) {
+	zonelist_type* next = zlt->next;
         free((void*) zlt->name);
         if (zlt->dname) {
             ldns_rdf_deep_free(zlt->dname);
         }
         free((void*) zlt->input_file);
         free((void*) zlt);
+	zlt = next;
     }
 }
 
@@ -165,7 +221,6 @@ free_config(config_type* cfg)
         free_zonelist(cfg->zonelist);
         free_serverlist(cfg->serverlist);
         free_serverlist(cfg->notifylist);
-        ldns_resolver_free(cfg->xfrd);
         free((void*) cfg);
     }
 }
@@ -538,6 +593,9 @@ sig_handler(int sig)
         case SIGHUP:
             sig_quit = 1;
             break;
+	case SIGUSR1:
+	    sig_reload = 1;
+	    break;
         default:
             break;
     }
@@ -573,6 +631,7 @@ setup_daemon(config_type* config)
     sigfillset(&action.sa_mask);
     action.sa_flags = 0;
     sigaction(SIGHUP, &action, NULL);
+    sigaction(SIGUSR1, &action, NULL);
 
     pid = getpid();
     if (writepid(config->pidfile, pid) == -1)
@@ -827,7 +886,7 @@ free_sockets(sockets_type* sockets)
 }
 
 static int
-odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
+odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config, int kick_signer)
 {
     ldns_status status = LDNS_STATUS_OK;
     ldns_rr* axfr_rr = NULL, *soa_rr = NULL;
@@ -839,6 +898,7 @@ odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
     char dest_file[MAXPATHLEN];
     char engine_sign_cmd[MAXPATHLEN + 1024]; 
     int soa_seen = 0;
+    ldns_resolver* xfrd = NULL;
 
     /* soa serial query */
     if (!zone || !zone->dname) {
@@ -855,11 +915,21 @@ odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
             "Aborting AXFR");
         return -1;
     }
-    status = ldns_resolver_send_pkt(&apkt, config->xfrd, qpkt);
+    
+    /* Initialise LDNS resolver for AXFR */
+    xfrd = init_xfrd(config);
+
+    if (!xfrd) {
+	    log_msg(LOG_ERR, "Failed to initialise AXFR structure");
+	    return -1;
+    }
+
+    status = ldns_resolver_send_pkt(&apkt, xfrd, qpkt);
     if (status != LDNS_STATUS_OK) {
         log_msg(LOG_ERR, "zone fetcher failed to send SOA query: %s",
             ldns_get_errorstr_by_id(status));
         ldns_pkt_free(qpkt);
+	ldns_resolver_free(xfrd);
         return -1;
     }
     if (ldns_pkt_ancount(apkt) == 1) {
@@ -874,15 +944,17 @@ odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
             "Aborting AXFR");
         /* retry? */
         ldns_pkt_free(apkt);
+	ldns_resolver_free(xfrd);
         return -1;
     }
 
     if (DNS_SERIAL_GT(new_serial, serial)) {
-        status = ldns_axfr_start(config->xfrd, zone->dname, LDNS_RR_CLASS_IN);
+        status = ldns_axfr_start(xfrd, zone->dname, LDNS_RR_CLASS_IN);
         ldns_pkt_free(qpkt);
         if (status != LDNS_STATUS_OK) {
             log_msg(LOG_ERR, "zone fetcher failed to start axfr: %s",
                 ldns_get_errorstr_by_id(status));
+	    ldns_resolver_free(xfrd);
             return -1;
         }
 
@@ -898,17 +970,19 @@ odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
             if (!fd) {
                 log_msg(LOG_ERR, "zone fetcher cannot store AXFR to file %s",
                     axfr_file);
+	        ldns_resolver_free(xfrd);
                 return -1;
             }
         }
 
         assert(fd);
 
-        axfr_rr = ldns_axfr_next(config->xfrd);
+        axfr_rr = ldns_axfr_next(xfrd);
         if (!axfr_rr) {
             log_msg(LOG_ERR, "zone fetcher AXFR for %s failed", zone->name);
 	    fclose(fd);
             unlink(axfr_file);
+	    ldns_resolver_free(xfrd);
             return -1;
         }
         else {
@@ -922,8 +996,26 @@ odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
                     ldns_rr_print(fd, axfr_rr);
                 }
                 ldns_rr_free(axfr_rr);
-                axfr_rr = ldns_axfr_next(config->xfrd);
+                axfr_rr = ldns_axfr_next(xfrd);
             }
+
+            /* RoRi:
+	     * We MUST now check if the AXFR was successful by verifying that
+	     * LDNS has seen the SOA record twice. Not doing this can result
+	     * in a half-transferred zone if the AXFR is interrupted.
+	     */
+	    if (!ldns_axfr_complete(xfrd)) {
+		 /* The AXFR was not successful, we've received only a partial zone */
+		 log_msg(LOG_ERR, "zone fetcher AXFR for %s failed, received only a partial zone", zone->name);
+		 fclose(fd);
+		 unlink(axfr_file);
+		 ldns_resolver_free(xfrd);
+		 return -1;
+	    }
+
+	    ldns_resolver_free(xfrd);
+	    xfrd = NULL;
+
             log_msg(LOG_INFO, "zone fetcher transferred zone %s serial %u "
                 "successfully", zone->name, new_serial);
 
@@ -933,12 +1025,14 @@ odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
             /* moving and kicking */
             snprintf(dest_file, sizeof(dest_file), "%s.axfr", zone->input_file);
             if(rename(axfr_file, dest_file) == 0) {
-                snprintf(engine_sign_cmd, sizeof(engine_sign_cmd),
-                    "%s %s", SIGNER_CLI_SIGN, zone->name);
-                if (system(engine_sign_cmd) != 0) {
-                    log_msg(LOG_ERR, "zone fetcher could not kick "
-                        "the signer engine to sign zone %s", zone->name);
-                }
+		if (kick_signer) {
+                    snprintf(engine_sign_cmd, sizeof(engine_sign_cmd),
+                        "%s %s", SIGNER_CLI_SIGN, zone->name);
+                    if (system(engine_sign_cmd) != 0) {
+                        log_msg(LOG_ERR, "zone fetcher could not kick "
+                            "the signer engine to sign zone %s", zone->name);
+                    }
+		}
             }
             else {
                 log_msg(LOG_ERR, "zone fetcher could not move AXFR to %s",
@@ -955,58 +1049,6 @@ odd_xfer(zonelist_type* zone, uint32_t serial, config_type* config)
     return 0;
 }
 
-static ldns_resolver*
-init_xfrd(config_type* config)
-{
-    serverlist_type* servers;
-    ldns_rdf* ns = NULL;
-    ldns_status status = LDNS_STATUS_OK;
-
-    ldns_resolver* xfrd = ldns_resolver_new();
-    if (config) {
-        if (config->use_tsig) {
-            ldns_resolver_set_tsig_keyname(xfrd, config->tsig_name);
-            if (strncmp(config->tsig_algo, "hmac-md5", 8) == 0) {
-                ldns_resolver_set_tsig_algorithm(xfrd, "hmac-md5.sig-alg.reg.int.");
-            } else {
-                ldns_resolver_set_tsig_algorithm(xfrd, config->tsig_algo);
-            }
-            ldns_resolver_set_tsig_keydata(xfrd, config->tsig_secret);
-        }
-        if (config->serverlist && config->serverlist->port)
-            ldns_resolver_set_port(xfrd, atoi(config->serverlist->port));
-        else
-            ldns_resolver_set_port(xfrd, atoi(DNS_PORT_STRING));
-        ldns_resolver_set_recursive(xfrd, 0);
-
-        servers = config->serverlist;
-        while (servers) {
-            if (servers->family == AF_INET6)
-                ns = ldns_rdf_new_frm_str(LDNS_RDF_TYPE_AAAA, servers->ipaddr);
-            else
-                ns = ldns_rdf_new_frm_str(LDNS_RDF_TYPE_A, servers->ipaddr);
-            if (ns)
-                status = ldns_resolver_push_nameserver(xfrd, ns);
-            else {
-                log_msg(LOG_ERR, "zone fetcher could not use %s for transfer "
-                    "request: could not parse ip address", servers->ipaddr);
-            }
-            if (status != LDNS_STATUS_OK) {
-                log_msg(LOG_ERR, "zone fetcher could not use %s for transfer "
-                    "request: %s", servers->ipaddr,
-                    ldns_get_errorstr_by_id(status));
-            }
-            servers = servers->next;
-        }
-        if (ldns_resolver_nameserver_count(xfrd) <= 0) {
-            log_msg(LOG_ERR, "zone fetcher could not find any valid name "
-                "servers");
-        }
-
-    }
-    return xfrd;
-}
-
 static void
 send_udp(uint8_t* buf, size_t len, void* data)
 {
@@ -1120,7 +1162,7 @@ handle_query(uint8_t* inbuf, ssize_t inlen,
                 serial = lookup_serial(fd);
                 fclose(fd);
             }
-            if (odd_xfer(zonelist, serial, config) != 0) {
+            if (odd_xfer(zonelist, serial, config, 1) != 0) {
                 log_msg(LOG_ERR, "AXFR for zone '%s' failed", zonelist->name);
             }
             ldns_pkt_free(query_pkt);
@@ -1277,6 +1319,73 @@ handle_tcp(int tcp_sock, config_type* config)
     close(s);
 }
 
+/* Reload the zonelist file and merge it with the existing configuration */
+static void reload_zonelist (config_type *config) {
+    zonelist_type *new_zonelist, **thisp;
+    zonelist_type *added_zonelist = NULL, *kept_zonelist = NULL;
+    int added_count = 0, changed_count = 0, kept_count = 0;
+    /* Fail softly if the zonelist cannot be accessed for reloading */
+    if (!zonelist_file) {
+	log_msg(LOG_ERR, "zone fetcher is unable to access the zonelist");
+	return;
+    } else {
+	log_msg(LOG_INFO, "zone fetcher will reload the zonelist");
+    }
+    /* Read the zonelist file and construct a new linked list of zonelist entries */
+    new_zonelist = read_zonelist (zonelist_file);
+    /* Iterate over the new zonelist file and compare it to previously configured zonelist entries */
+    while (new_zonelist) {
+	zonelist_type *next_zonelist = new_zonelist->next;
+	zonelist_type *this = config->zonelist;
+	int found = 0;
+	while (this && !found) {
+	   found = !strcmp (this->name, new_zonelist->name);
+	   if (!found)
+		this = this->next;
+	}
+	/* If the zone name is found in the old zonelist, it is either a full match or a replacement */
+	if (found) {
+	    if (strcmp (new_zonelist->input_file, this->input_file)) {
+		/* the zonelist entry has changed -- treat as a replacement/new zonelist entry */
+		changed_count++;
+		new_zonelist->next = added_zonelist;
+		added_zonelist = new_zonelist;
+	    } else {
+		/* the zonelist entry is already configured -- treat as a kept zonelist entry */
+		kept_count++;
+		new_zonelist->next = kept_zonelist;
+		kept_zonelist = new_zonelist;
+	    }
+	} else {
+	    /* new_zonelist introduces a new zonelist entry */
+	    added_count++;
+	    new_zonelist->next = added_zonelist;
+	    added_zonelist = new_zonelist;
+	}
+	new_zonelist = next_zonelist;
+    }
+    /* Replace the configured zonelist with the added_zonelist and kept_zonelist */
+    free_zonelist (config->zonelist);
+    config->zonelist = kept_zonelist;
+    thisp = &config->zonelist;
+    while (*thisp) {
+	thisp = &(*thisp)->next;
+    }
+    *thisp = added_zonelist;
+    /* Perform an initial AXFR for the newly added zones (assume no present inputfile) */
+    new_zonelist = added_zonelist;
+    while (new_zonelist) {
+	/* send the request -- assume no file is present so SOA is 0 */
+	if (odd_xfer (new_zonelist, 0, config, 0) != 0) {
+	    log_msg(LOG_ERR, "AXFR for new zone '%s' failed", new_zonelist->name);
+	}
+	/* next */
+	new_zonelist = new_zonelist->next;
+    }
+    log_msg(LOG_INFO, "Reloaded zonelist -- kept %d, changed %d and added %d zones",
+		kept_count, changed_count, added_count);
+}
+
 static void
 xfrd_ns(sockets_type* sockets, config_type* cfg)
 {
@@ -1289,7 +1398,11 @@ xfrd_ns(sockets_type* sockets, config_type* cfg)
     count = 0;
     timeout.tv_sec = 0;
     timeout.tv_usec = 0;
-    while (1) {
+    while (!sig_quit) {
+	if (sig_reload) {
+	    reload_zonelist (cfg);
+	    sig_reload = 0;
+	}
         FD_ZERO(&rset);
         FD_ZERO(&wset);
         FD_ZERO(&eset);
@@ -1304,7 +1417,7 @@ xfrd_ns(sockets_type* sockets, config_type* cfg)
 
         if (select(maxfd+1, &rset, &wset, &eset, NULL) < 0) {
             if (errno == EINTR)
-                break;
+                continue;
             log_msg(LOG_ERR, "zone fetcher select(): %s", strerror(errno));
         }
 
@@ -1351,12 +1464,6 @@ list_settings(config_type* config, const char* filename)
             fprintf(stdout, "\t%s\n", servers->ipaddr);
             servers = servers->next;
         }
-        if (config->xfrd) {
-            fprintf(stdout, "transfer daemon available\n");
-        }
-        else {
-            fprintf(stdout, "transfer daemon missing\n");
-        }
     }
     else fprintf(stdout, "no config\n");
 }
@@ -1364,7 +1471,7 @@ list_settings(config_type* config, const char* filename)
 int
 main(int argc, char **argv)
 {
-    const char* zonelist_file = NULL, *config_file = NULL;
+    const char* config_file = NULL;
     zonelist_type *zonelist = NULL;
     config_type* config = NULL;
     int c, run_as_daemon = 0;
@@ -1431,8 +1538,7 @@ main(int argc, char **argv)
     config->pidfile = strdup(PID_FILENAME_STRING); /* not freed */
     c = read_axfr_config(config_file, config);
     config->zonelist = read_zonelist(zonelist_file);
-    config->xfrd = init_xfrd(config);
-	xmlCleanupParser();
+    xmlCleanupParser();
 
     if (info) {
         list_settings(config, config_file);
@@ -1461,14 +1567,14 @@ main(int argc, char **argv)
             fclose(fd);
         }
         /* send the request */
-        if (odd_xfer(zonelist, serial, config) != 0) {
+        if (odd_xfer(zonelist, serial, config, 1) != 0) {
             log_msg(LOG_ERR, "AXFR for zone '%s' failed", zonelist->name);
         }
         /* next */
         zonelist = zonelist->next;
     }
 
-	if (run_as_daemon) {
+    if (run_as_daemon) {
         pid = setup_daemon(config);
         /* listen to NOTIFY messages */
         c = init_sockets(&sockets, config->notifylist);
@@ -1492,6 +1598,22 @@ main(int argc, char **argv)
             log_msg(LOG_NOTICE, "zone fetcher exiting...");
             exit(EXIT_FAILURE);
         }
+	/* if chroot used, strip it off zonelist_file */
+	if (chroot && zonelist_file) {
+	    size_t chrootlen = strlen (chroot);
+	    if ((chrootlen > 0) && (chroot [chrootlen-1] == '/'))
+		chrootlen--;
+	    if ((strlen (zonelist_file) <= chrootlen)
+			|| strncmp (zonelist_file, chroot, chrootlen)
+			|| (zonelist_file [chrootlen] != '/')) {
+		log_msg(LOG_ERR, "zone fetcher will not be able to reload zonelist after chroot");
+		zonelist_file = NULL;
+	    } else {
+		char *new_zonelist_file = strdup (zonelist_file + chrootlen);
+		free (zonelist_file);
+		zonelist_file = new_zonelist_file;
+	    }
+	}
 
         xfrd_ns(&sockets, config);
 
